unit clgrep;
{
        clgrep.pas -- Container Lite v 1.91
        (C) Copyright 1996  John Webster Small
        All rights reserved
}

interface

uses SysUtils, Contain, Cstring;

const CL_GREP_MAX_RECUR = 6;
      NDEBUG_GRE : boolean = TRUE;
      NTRACE_EXP_MATCH : boolean = TRUE;
      NTRACE_EXP_MATCH_SEMA4_ON : word = 0;
      NTRACE_EXP_MATCH_SEMA4_OFF : word = 1;
      NTRACE_TOK_MATCH : boolean = TRUE;
      NTRACE_GROUP_MATCH : boolean = FALSE;

procedure MATCH_TRACE_ON;
procedure MATCH_TRACE_OFF;

type

    EgrepException = class(Exception);

    grep_exp = class(Container)
      {  Helper class to grep_subExp and CL_grep }
        fGRE : cstr;
        procedure GRE_cpy;
      public
        constructor Create;
        destructor Destroy; override;
        function parse(const GRE : PChar; GRE_len : word;
            groups : boolean) : boolean;
        procedure parseGroup
            (leftDelimiter, rightDelimiter : char;
            const GRE : PChar; GRE_len : word;
            var tokLen : word; groups : boolean);
        function match(const SS : PChar;
            var len : word; caseIgnore : boolean;
            groupToks : Container) : boolean;
        procedure traceMatch(msg, SS : PChar);
        property pGRE : cstr read fGRE;
    end;

    CL_grep = class
      private
        parseTree : grep_exp;
        gre_cstr : cstr;
        BOL_anchor, EOL_anchor, _groups : boolean;
        fPrefix, fToken, fSuffix : cstr;
      protected
        function _init(const GRE : PChar;
            groups : boolean) : boolean;
        constructor _Create_specs
            (const GRE : PChar; groups : boolean);
        function _match(const SS : PChar;
            var idx, len : word; caseIgnore : boolean;
            groupToks : Container) : boolean;
      public
        constructor Create_specs(const GRE : PChar);
        constructor Create;
        destructor Destroy; override;
        function init(const GRE : PChar) : boolean; virtual;
        function grepOk : boolean;
        function match(const SS : PChar; var idx, len : word;
            caseIgnore : boolean) : boolean; virtual;
        function parse_specs
            (const SS : PChar; caseIgnore : boolean) : boolean;
        function parse(const SS : PChar) : boolean;
        property Prefix : cstr read fPrefix;
        property Token : cstr read fToken;
        property Suffix : cstr read fSuffix;
        procedure clr;
    end;

    CL_ggrep = class(CL_grep)
      private
        ftokens : Container;
      public
        property tokens : Container read ftokens;
        constructor Create_specs(const GRE : PChar);
        constructor Create;
        destructor Destroy; override;
        function init(const GRE : PChar) : boolean; override;
        function groupToksOk : boolean;
        function match(const SS : PChar; var idx, len : word;
            caseIgnore : boolean) : boolean; override;
        function parse_specs
            (const SS : PChar; caseIgnore : boolean) : boolean;
        function parse(const SS : PChar) : boolean;
        function filter_specs
            (const SS : PChar; caseIgnore : boolean;
             Mask : PChar; Generated : cstr) : boolean;
        function filter(const SS, Mask : PChar; Generated : cstr) : boolean;
    end;

implementation

uses ctype;

const traceFileName : string = '\clgrep.log';

var grep_exp_sema4 : word;
    NTRACE : boolean;
    traceCstr : cstr;
    traceFile : TextFile;
    ExitSave : Pointer;

procedure MATCH_TRACE_ON;
begin
  if NTRACE then begin
     NTRACE := FALSE;
     traceCstr := cstr.Create;
     AssignFile(traceFile,traceFileName);
     try
       Append(traceFile);
       writeln(traceFile,'******************************');
     except
       Rewrite(traceFile);
     end;
     end;
end;

procedure MATCH_TRACE_OFF;
begin
  if not NTRACE then begin
     NTRACE := TRUE;
     traceCstr.Free;
     traceCstr := nil;
     CloseFile(traceFile);
     end;
end;

procedure traceWriteln;
begin
  traceCstr.printOnFile(traceFile);
  writeln(traceFile);
end;

type

    occurences = ( ZeroOrOne, One, OneOrMore, Any );

    grep_tok = class
      { plays double duty as polycluster root
        and the "." wildcard }
      protected
        fGRE : cstr;
        procedure GRE_catCount;
        procedure GRE_cpy; virtual;
      public
        count: occurences;
        constructor Create;
        destructor Destroy; override;
        function matchOne
            (const SS : PChar; var len : word;
             caseIgnore : boolean;
             groupToks : Container) { container of cstr }
             : boolean; virtual;
        class function newTok(const GRE : PChar;
            GRE_len : word; var tokLen : word;
            groups : boolean) : grep_tok;
        class procedure error(const GRE : PChar;
            pos : word; const msg : PChar);
        procedure traceMatchOne(SS : PChar; matched : boolean);
        property pGRE : cstr read fGRE;
    end;


    grep_subExp = class(grep_tok)
      private
        fexp : grep_exp;
      protected
        procedure GRE_cpy; override;
      public
        constructor Create(const GRE : PChar;
            GRE_len : word; var tokLen : word;
            groups : boolean);
        destructor Destroy; override;
        function matchOne(const SS : PChar;
            var len : word; caseIgnore : boolean;
            groupToks : Container) : boolean; override;
    end;

    grep_group = class(grep_tok)
      private
        fexp : grep_exp;
      protected
        procedure GRE_cpy; override;
      public
        constructor Create(const GRE : PChar;
            GRE_len : word; var tokLen : word);
        destructor Destroy; override;
        function matchOne(const SS : PChar;
            var len : word; caseIgnore : boolean;
            groupToks : Container) : boolean; override;
        procedure debugGroupMatch(GT : cstr);
    end;


    grep_seq = class(grep_tok)
      private
        fseq : cstr;
      protected
        procedure GRE_cpy; override;
      public
        constructor Create(const GRE : PChar;
            GRE_len : word; var tokLen : word;
            groups : boolean);
        destructor Destroy; override;
        function matchOne(const SS : PChar;
            var len_ : word; caseIgnore : boolean;
            groupToks : Container) : boolean; override;
    end;

    grep_setTok = class
      protected
        procedure GRE_cat(GRE : cstr); virtual;
      public
        constructor Create;
        function match(c : char; caseIgnore : boolean)
            : boolean; virtual;
    end;

    grep_setAlts = class(grep_setTok)
      private
        fseq : cstr;
      protected
        procedure GRE_cat(GRE : cstr); override;
      public
        constructor Create(const SEQ : PChar);
        destructor Destroy; override;
        function match(c : char; caseIgnore : boolean)
            : boolean; override;
    end;

    grep_setRange = class(grep_setTok)
      private
        _left, _right : char;
      protected
        procedure GRE_cat(GRE : cstr); override;
      public
        constructor Create(left, right : char);
        function match(c : char; caseIgnore : boolean)
            : boolean; override;
    end;

    grep_set = class(grep_tok)
      private
        fsetToks : Container;
        negate : boolean;
      protected
        procedure GRE_cpy; override;
      public
        constructor Create(const GRE : PChar;
            GRE_len : word; var tokLen : word);
        destructor Destroy; override;
        function matchOne(const SS : PChar;
            var len : word; caseIgnore : boolean;
            groupToks : Container) : boolean; override;
    end;

    grep_alt = class(grep_tok)
      private
        _left, _right : grep_tok;
      protected
        procedure GRE_cpy; override;
      public
        constructor Create(const GRE : PChar;
            GRE_len : word; leftTok : grep_tok;
            var rightTokLen : word; groups : boolean);
        destructor Destroy; override;
        function matchOne(const SS : PChar;
            var len : word; caseIgnore : boolean;
            groupToks : Container) : boolean; override;
    end;

procedure grep_exp.GRE_cpy;
var i : word;
    T : grep_tok;
begin
  if (not Assigned(fGRE)) then exit;
  i := 0;
  fGRE.clr;
  while i < nodes do begin
    T := grep_tok(atGet(i));
    fGRE.cat(T.pGRE.Str);
    inc(i);
    end;
end;

constructor grep_exp.Create;
begin
  inherited Create_specs(grep_tok,CL_ALL_DESCENDANTS,CL_FREE);
  if NDEBUG_GRE then
     fGRE := nil
  else
     fGRE := cstr.Create;
end;

destructor grep_exp.Destroy;
begin
  fGRE.Free;
  inherited Destroy;
end;

function grep_exp.parse(const GRE : PChar; GRE_len : word;
    groups : boolean) : boolean;
var tokLen, i : word;
begin
    result := false;
    if grep_exp_sema4 >= CL_GREP_MAX_RECUR then
       raise EgrepException.Create('GREP nesting excessive.');
    inc(grep_exp_sema4);
    tokLen := 0;
    if (GRE = nil) or (GRE_len = 0) then
       raise EgrepException.Create('Regular expression missing.');
    i := 0;
    while i < GRE_len do begin
       if (not insQFL(grep_tok.
            newTok(@(GRE[i]),GRE_len-i,tokLen,groups))) then begin
            allFree;
            dec(grep_exp_sema4);
            exit;
            end;
       inc(i,tokLen);
       end;
    dec(grep_exp_sema4);
    result := true;
    GRE_cpy;
end;

procedure grep_exp.parseGroup(leftDelimiter, rightDelimiter : char;
    const GRE : PChar; GRE_len : word;
    var tokLen : word; groups : boolean);
var nestcspn : string[4];
    cspnLen, oldLen : word;
    group : cstr;
    nesting : integer;
begin
    nestcspn := leftDelimiter + rightDelimiter + '\' + #0;
    if (GRE = nil) or (GRE[0] <> leftDelimiter) then
       raise EgrepException.Create('Grouping error');
    tokLen := 1;
    group := cstr.Create;
    if not Assigned(group) then exit;
  try
    nesting := 1;
    while (tokLen < GRE_len) do begin
        cspnLen := strcspn(@(GRE[tokLen]),@(nestcspn[1]));
        oldLen := group.len;
        group.ncat_extra(@(GRE[tokLen]),cspnLen,2);
        if group.len <> oldLen + cspnLen then
           break;
        inc(tokLen,cspnLen);
        if (tokLen >= GRE_len) then begin
            grep_tok.error(GRE,tokLen,
                'nesting unbalanced');
            break;
            end;
        if (GRE[tokLen] = '\')  then begin
            if (tokLen + 1 >= GRE_len) then begin
                grep_tok.error
                    (GRE,tokLen+1,'nothing after \');
                break;
                end;
            group.ncat(@(GRE[tokLen]),2);
            inc(tokLen,2);
            end
        else if (GRE[tokLen] = leftDelimiter) then begin
            group.ncat(@(GRE[tokLen]),1);
            inc(tokLen);
            inc(nesting);
            end
        else begin  (*  found ")" or "}" *)
            dec(nesting);
            if (nesting <> 0) then begin
                group.ncat(@(GRE[tokLen]),1);
                inc(tokLen);
                end
            else  begin
                inc(tokLen);
                if (parse(group.str,group.len,
                    groups)) then begin
                    exit;
                    end;
            end
        end
    end;
    tokLen := 0;
  finally
    group.Free
  end;
end;

function grep_exp.match(const SS : PChar;
    var len : word; caseIgnore : boolean;
    groupToks : Container) : boolean;
var SSlen, tokLen, i, maxLen : word;
    T, nextTok : grep_tok;
begin
{ set NDEBUG_EXP = False before creating CL_grep/CL_ggrep
  then watch SS, fGRE._S, PChar(@SS[len]), T.fGRE._S,
  PChar(@SS[maxLen]), nextTok.fGRE._S
}
    result := false;
    if SS = nil then
       SSlen := 0
    else
       SSlen := strlen(SS);
    len := 0;
    i := 0;
    T := atGet(i) as grep_tok;
    while (T <> nil) and (len < SSlen) do begin
        if (not T.matchOne(@(SS[len]),tokLen,
           caseIgnore,groupToks)) then begin
           if (T.count in [ZeroOrOne,Any]) then begin
              tokLen := 0;
              inc(len,tokLen);
              inc(i);
              T := atGet(i) as grep_tok;
              continue;
              end;
           exit; { OneOrMore, One }
           end;
        if (T.count in [One,ZeroOrOne]) then begin
            inc(len,tokLen);
            inc(i);
            T := atGet(i) as grep_tok;
            continue;
            end;
        { Searching for OneOrMore, Any }
        maxLen := len + tokLen;
        nextTok := atGet(i + 1) as grep_tok;
        while (maxLen < SSlen) do begin
            if (nextTok <> nil) then   { minimal matching }
             { Look ahead before matching more of same }
                if nextTok.matchOne(@(SS[maxLen]),tokLen,
                    caseIgnore,nil) then
                    break;
            if (not T.matchOne
                (@(SS[maxLen]),tokLen,
                caseIgnore,groupToks)) then
                break;
            inc(maxLen,tokLen);
        end;
        tokLen := maxLen - len;
        inc(i);
        inc(len,tokLen);
        T := atGet(i) as grep_tok;
    end;

    while (T <> nil) do
      if (T.count in [ZeroOrOne,Any]) then begin
         inc(i);
         T := atGet(i) as grep_tok;
         end
      else
         break;
    if (T = nil) then begin
       result := true;
       traceMatch('matched: ',SS)
       end;
end;

procedure grep_exp.traceMatch(msg, SS : PChar);
begin
  if NTRACE then exit;
  if NTRACE_EXP_MATCH then exit;
  if grep_exp_sema4 >= NTRACE_EXP_MATCH_SEMA4_OFF then exit;
  if grep_exp_sema4 > 0 then
     if grep_exp_sema4 < NTRACE_EXP_MATCH_SEMA4_ON then exit;
  if SS = nil then
     raise EgrepException.Create('Search string missing!');
  traceCstr.cpy(msg).cat(pGRE.str).cat(' @ ').cat(SS);
  traceWriteln;
end;

function CL_grep._init(const GRE : PChar;
    groups : boolean) : boolean;
var GRE_len, i : word;
    tmp : cstr;
begin
    result := false;
    if not Assigned(parseTree) or not Assigned(gre_cstr) then
       exit;
    parseTree.allFree;
    gre_cstr.trunc(0);
    BOL_anchor := false;
    EOL_anchor := false;
    _groups := groups;
    fPrefix.clr; fToken.clr; fSuffix.clr;
    if GRE <> nil then begin
      GRE_len := strlen(GRE);
      if (GRE_len <> 0) then begin
        i := 0;
        { look for BOL, EOL pattern anchors }
        if (GRE[i] = '^')  then begin
            BOL_anchor := true;
            inc(i);
        end;
        if (GRE[GRE_len - 1] = '$') then begin
            EOL_anchor := true;
            dec(GRE_len);
        end;
        { parse tokens }
        tmp := cstr.Create;
        try
          tmp.ncpy(@(GRE[i]),GRE_len-i);
          if tmp.len = GRE_len-i then
            if (parseTree.parse
              (tmp.str,tmp.len,groups)) then begin
              { GRE valid, parse tree complete }
              gre_cstr.cpy(GRE);
              result := true;
            end;
        finally
          tmp.Free;
        end;
      end;
    end;
end;

constructor CL_grep._Create_specs
    (const GRE : PChar; groups : boolean);
begin
  inherited Create;
  parseTree := grep_exp.Create;
  gre_cstr := cstr.Create;
  BOL_anchor := false;
  EOL_anchor := false;
  _groups := false;
  fPrefix := cstr.Create;
  fToken := cstr.Create;
  fSuffix := cstr.Create;
  _init(GRE,groups);
end;

function CL_grep._match(const SS : PChar;
    var idx, len : word; caseIgnore : boolean;
    groupToks : Container) : boolean;
var SSlen : word;
begin
  result := false;
  fPrefix.clr; fToken.clr; fSuffix.clr;
  if Assigned(groupToks) then begin
     if not _groups then
        raise EgrepException.Create('groupToks inappropriate');
     if groupToks.rootType <> cstr then
        raise EgrepException.Create('groupToks must be Container of cstr');
     groupToks.allClr;
     groupToks.setMaxNodes(CL_MAXNODES);
     groupToks.disposableNodes := true;
  end;
  if SS = nil then
     SSlen := 0
  else
     SSlen := strlen(SS);
  if not Assigned(parseTree) then
     raise EgrepException.Create('Parse tree missing');
  if (parseTree.nodes > 0) and (SSlen > 0) then begin
    idx := 0;
    parseTree.traceMatch('Attempt: ',SS);
    while idx < SSlen do begin
      if (parseTree.match(@(SS[idx]),len,
          caseIgnore,groupToks)) then
          if (not EOL_anchor) or (idx + len >= SSlen) then begin
              result := true;
              exit;
              end;
      if Assigned(groupToks) then
         groupToks.AllClr;
      if (BOL_anchor) then
          exit;
      inc(idx);
    end;
  end;
end;

constructor CL_grep.Create_specs(const GRE : PChar);
begin
  inherited Create;
  parseTree := grep_exp.Create;
  gre_cstr := cstr.Create;
  BOL_anchor := false;
  EOL_anchor := false;
  _groups := false;
  fPrefix := cstr.Create;
  fToken := cstr.Create;
  fSuffix := cstr.Create;
  _init(GRE,false);
end;

constructor CL_grep.Create;
begin
  inherited Create;
  parseTree := grep_exp.Create;
  gre_cstr := cstr.Create;
  BOL_anchor := false;
  EOL_anchor := false;
  _groups := false;
  fPrefix := cstr.Create;
  fToken := cstr.Create;
  fSuffix := cstr.Create;
  _init(nil,false);
end;

destructor CL_grep.Destroy;
begin
  parseTree.Free;
  gre_cstr.Free;
  fPrefix.Free;
  fToken.Free;
  fSuffix.Free;
  inherited Destroy;
end;

function CL_grep.init(const GRE : PChar) : boolean;
  begin result := _init(GRE,false); end;

function CL_grep.grepOk : boolean;
  begin result := parseTree.nodes > 0; end;

function CL_grep.match(const SS : PChar;
    var idx, len : word; caseIgnore : boolean)
    : boolean;
begin
  result := _match(SS,idx,len,caseIgnore,nil);
  if result then begin
     fPrefix.ncpy(SS,idx);
     fToken.ncpy(@SS[idx],len);
     fSuffix.cpy(@SS[idx+len]);
     end;
end;

function CL_grep.parse_specs
  (const SS : PChar; caseIgnore : boolean) : boolean;
var idx, len : word;
begin
  result := match(SS,idx,len,caseIgnore);
end;

function CL_grep.parse(const SS : PChar) : boolean;
begin
  result := parse_specs(SS,false);
end;

procedure CL_grep.clr;
begin
  fPrefix.clr;
  fToken.clr;
  fSuffix.clr;
end;

constructor CL_ggrep.Create_specs(const GRE : PChar);
begin
  inherited _Create_specs(GRE,true);
  ftokens := Container.Create_specs(cstr,CL_ALL_DESCENDANTS,CL_FREE);
end;

constructor CL_ggrep.Create;
begin
  inherited _Create_specs(nil,true);
  ftokens := Container.Create_specs(cstr,CL_ALL_DESCENDANTS,CL_FREE);
end;

destructor CL_ggrep.Destroy;
begin
  ftokens.Free;
  inherited Destroy;
end;

function CL_ggrep.init(const GRE : PChar) : boolean;
begin
    if Assigned(ftokens) then begin
       ftokens.allClr;
       result := _init(GRE,true);
       end
    else
       result := false;
end;

function CL_ggrep.groupToksOk : boolean;
begin
    if Assigned(ftokens) then
       result := (ftokens.maxNodes = CL_MAXNODES)
    else
       result := false;
end;

function CL_ggrep.match(const SS : PChar;
    var idx, len : word; caseIgnore : boolean)
    : boolean;
begin
    result := _match(SS,idx,len,caseIgnore,ftokens);
    if result and (ftokens.nodes > 0) then
       Token.cpy(cstr(ftokens[0]).str);
end;

function CL_ggrep.parse_specs
    (const SS : PChar; caseIgnore : boolean) : boolean;
var idx, len : word;
begin
    result := match(SS,idx,len,caseIgnore);
end;

function CL_ggrep.parse(const SS : PChar) : boolean;
begin
    result := parse_specs(SS,false);
end;

function CL_ggrep.filter_specs(const SS : PChar; caseIgnore : boolean;
  Mask : PChar; Generated : cstr) : boolean;
var idx, len, i : word;
    tok : cstr;
    MaskParser : CL_grep;
begin
  result := false;
  if (not Assigned(Generated)) or (SS = nil) or (Mask = nil) then exit;
  Generated.clr;
  result := parse_specs(SS,caseIgnore);
  if result then begin
     Generated.cpy(Mask);
     MaskParser := CL_grep.Create_specs('\\[1-9]');
     try
     while MaskParser.match(Generated.Str,idx,len,false) do begin
       i := Ord(MaskParser.token[2])-Ord('1');  { 0 indexed }
       if i < tokens.nodes then
          Generated.replace_PChar_fill(idx,len,cstr(tokens[i]).str)
       else
          Generated.replace_PChar_fill(idx,len,'');
       end;
     finally
       MaskParser.Free;
     end;
     end;
end;

function CL_ggrep.filter(const SS, Mask : PChar; Generated : cstr) : boolean;
begin result := filter_specs(SS,false,Mask,Generated); end;

procedure grep_tok.GRE_catCount;
begin
  if not Assigned(fGRE) then exit;
  case count of
    ZeroOrOne: fGRE.cat('?');
    OneOrMore: fGRE.cat('+');
    Any: fGRE.cat('*');
  end;
end;

procedure grep_tok.GRE_cpy;
begin
  if not Assigned(fGRE) then exit;
  fGRE.cpy('.');
  GRE_catCount;
end;

constructor grep_tok.Create;
begin
  inherited Create;
  count := One;
  if NDEBUG_GRE then
     fGRE := nil
  else
     fGRE := cstr.Create;
end;

destructor grep_tok.Destroy;
begin
  fGRE.Free;
  inherited Destroy;
end;

function grep_tok.matchOne
    (const SS : PChar; var len : word;
    caseIgnore : boolean; groupToks : Container)
    : boolean;
begin
    if (not Assigned(SS)) or (strlen(SS) = 0) then
        raise EgrepException.Create('Search string missing.');
    len := 1;
    result := true;
    traceMatchOne(SS,result);
end;

class function grep_tok.newTok(const GRE : PChar;
     GRE_len : word; var tokLen : word;
     groups : boolean) : grep_tok;
var T, oldT : grep_tok;
    rightTokLen : word;
begin
   { GRE has already been stripped of leading ^ and trailing $. }
    result := nil;
    tokLen :=0;
    if not Assigned(GRE) then
        raise EgrepException.Create('Regular expression missing.');
    if GRE_len = 0 then
        exit;
    if (strspn(GRE,')]|*+?') <> 0) or
        (groups and (GRE[0] = '}')) then begin
        error(GRE,0,'illegal grep char');
        exit;
        end;
    T := nil;
    oldT := nil;
  try
    case GRE[0] of
        '[': T := grep_set.Create(GRE,GRE_len,tokLen);
        '(': T := grep_subExp.Create(GRE,GRE_len,tokLen,groups);
        '.': begin T := grep_tok.Create; tokLen := 1; end;
        '{': if groups then
                T := grep_group.Create(GRE,GRE_len,tokLen)
            { else '{' is like any other character }
             else
                T := grep_seq.Create(GRE,GRE_len,tokLen,groups);
        else
            T := grep_seq.Create(GRE,GRE_len,tokLen,groups);
        end;
    if Assigned(T) then if (tokLen = 0) then begin
        T.Free;
        T := nil;
        end
    else  begin
        case (GRE[tokLen]) of
          '*' : begin inc(tokLen); T.count := Any; T.GRE_cpy; end;
          '+' : begin inc(tokLen); T.count := OneOrMore; T.GRE_cpy; end;
          '?' : begin inc(tokLen); T.count := ZeroOrOne; T.GRE_cpy; end;
          '|' : begin
                inc(tokLen);
                rightTokLen := 0;
                T.GRE_cpy;  { left side }
                oldT := T;
                T := grep_alt.Create(@(GRE[tokLen]),
                    GRE_len-tokLen,T,
                    rightTokLen,groups); { right size already  done }
                if Assigned(T) then  begin
                    inc(tokLen,rightTokLen);
                    T.GRE_cpy;  { alt top }
                    end
                else begin
                    tokLen := 0;
                    oldT.Free;
                    oldT := nil
                    end;
                end;
        else
          T.GRE_cpy;  { count = one }
        end;
        end;
    result := T;
  except
    T.Free;
    oldT.Free;
    raise;
  end
end;

class procedure grep_tok.error(const GRE : PChar;
             pos : word; const msg : PChar);
begin
    if GRE = nil then
       raise EgrepException.Create('Regular expression missing.');
    if msg = nil then
       raise EgrepException.Create(Format('grep error, pos: %d %s',[pos,GRE]))
    else
       raise EgrepException.Create(Format('%s pos: %d %s',[msg,pos,GRE]));
end;

procedure grep_tok.traceMatchOne(SS : PChar; matched : boolean);
begin
  if NTRACE then exit;
  if NTRACE_TOK_MATCH then exit;
  if SS = nil then
     raise EgrepException.Create('Search string missing!');
  traceCstr.ncpy_fmt(100,'%s.matchOne: ',[self.className]);
  if Assigned(fGRE) then
     traceCstr.cat(pGRE.str);
  traceCstr.cat(' @ ').cat(SS);
  if matched then
     traceCstr.cat('  !!MATCHED!!')
  else
     traceCstr.cat('  ~~FAILED~~');
  traceWriteln;
end;

procedure grep_subExp.GRE_cpy;
begin
  if (not Assigned(fexp)) or (not Assigned(fGRE)) then exit;
  pGRE.cpy('(');
  pGRE.cat(fexp.pGRE.str);
  pGRE.cat(')');
  GRE_catCount;
end;

constructor grep_subExp.Create(const GRE : PChar;
    GRE_len : word; var tokLen : word;
    groups : boolean);
begin
  inherited Create;
  fexp := grep_exp.Create;
  if Assigned(fexp) then
     fexp.parseGroup('(',')',GRE,GRE_len,tokLen,groups);
end;

destructor grep_subExp.Destroy;
begin
  fexp.Free;
  inherited Destroy;
end;

function grep_subExp.matchOne(const SS : PChar;
    var len : word; caseIgnore : boolean;
    groupToks : Container) : boolean;
begin
  if Assigned(fexp) then
     result := fexp.match(SS,len,caseIgnore,groupToks)
  else
     result := false;
  traceMatchOne(SS,result);
end;

procedure grep_group.GRE_cpy;
begin
  if (not Assigned(fexp)) or (not Assigned(fGRE)) then exit;
  pGRE.cpy('{');
  pGRE.cat(fexp.pGRE.str);
  pGRE.cat('}');
  GRE_catCount;
end;

constructor grep_group.Create(const GRE : PChar;
    GRE_len : word; var tokLen : word);
begin
  inherited Create;
  fexp := grep_exp.Create;
  if Assigned(fexp) then
     fexp.parseGroup('{','}',GRE,GRE_len,tokLen,true);
end;

destructor grep_group.Destroy;
begin
  fexp.Free;
  inherited Destroy;
end;

function grep_group.matchOne(const SS : PChar;
    var len : word; caseIgnore : boolean;
    groupToks : Container) : boolean;
var _GT : cstr;
begin
  result := false;
  if Assigned(fexp) then
   if (fexp.match(SS,len,caseIgnore,groupToks)) then begin
    if Assigned(groupToks) then begin
     if groupToks.rootType <> cstr then
       raise EgrepException.Create('groupToks must be Container of cstr');
     if groupToks.vacancy then begin
      _GT := cstr.Create_extra(len+1);
      if Assigned(_GT) then begin
        _GT.ncpy(SS,len);
        if (not groupToks.insQFL(_GT)) then
          groupToks.setMaxNodes(CL_MAXNODES-1)
        else
          debugGroupMatch(_GT);
      end
      else
        groupToks.setMaxNodes(CL_MAXNODES-1);
     end;
    end;
    result := true;
   end;
  traceMatchOne(SS,result);
end;

procedure grep_group.debugGroupMatch(GT : cstr);
begin
  if NTRACE then exit;
  if NTRACE_GROUP_MATCH then exit;
  traceCstr.clr;
  traceCstr.cpy('Group found: ');
  if Assigned(fGRE) then
     traceCstr.cat(pGRE.str);
  traceCstr.cat(' = ');
  traceCstr.cat(GT.str);
  traceWriteln;
end;

procedure grep_seq.GRE_cpy;
begin
  if (not Assigned(fseq)) or (not Assigned(fGRE)) then exit;
  pGRE.cpy(fseq.str);
  GRE_catCount;
end;

constructor grep_seq.Create(const GRE : PChar;
    GRE_len : word; var tokLen : word;
    groups : boolean);
var lastEscIdx, cspnLen, oldLen : word;
begin
    inherited Create;
    fseq := cstr.Create;
    if not Assigned(fseq) then exit;
    if GRE = nil then
       raise EgrepException.Create('Regular expression missing.');
    lastEscIdx := 0;
    tokLen := 0;
    while (tokLen < GRE_len)  do begin
        cspnLen := strcspn(@(GRE[tokLen]),'[()]*+?|\.{}');
        oldLen := fseq.len;
        fseq.ncat_extra(@(GRE[tokLen]),cspnLen,1);
        if fseq.len <> (oldLen + cspnLen) then begin
           tokLen := 0;
           exit;
           end;
        inc(tokLen,cspnLen);
        if (tokLen >= GRE_len) then
            break;
        if (strchr('{}',GRE[tokLen]) <> nil) then
            if (groups) then
                break  (* "{}"  *)
            else begin
                fseq.ncat(@(GRE[tokLen]),1);
                inc(tokLen);
                continue;
            end;
        if (GRE[tokLen] <> '\') then
            break;  (* '[()]*+?|.'  *)
        lastEscIdx := tokLen; { found '\' }
        inc(tokLen);
        if (tokLen < GRE_len) then begin
            fseq.ncat(@(GRE[tokLen]),1);
            inc(tokLen);
            end
        else begin
            error(GRE,tokLen,'nothing following \');
            tokLen := 0;
            exit;
        end;
    end;
    if (fseq.len > 1) and (tokLen < GRE_len) and
        (strspn(@(GRE[tokLen]),'*+?|') > 0)  then begin
        dec(tokLen);
        if (lastEscIdx = (tokLen - 1))
           and (GRE[lastEscIdx] = '\')  then
            dec(tokLen);
        fseq.trunc(fseq.len-1);
    end;
    fseq.pack;
    if (fseq.len = 0)  then
        tokLen := 0;
end;

destructor grep_seq.Destroy;
begin
  fseq.Free;
  inherited Destroy;
end;

function grep_seq.matchOne(const SS : PChar;
            var len_ : word; caseIgnore : boolean;
            groupToks : Container) : boolean;
begin
  len_ := 0;
  if Assigned(fseq) then
    if caseIgnore then begin
       if fseq.nicmp(SS,fseq.len) = 0 then
          len_ := fseq.len;
       end
    else
       if fseq.ncmp(SS,fseq.len) = 0 then
          len_ := fseq.len;
  result := len_ <> 0;
  traceMatchOne(SS,result);
end;

procedure grep_setTok.GRE_cat(GRE : cstr);
begin
  raise EgrepException.Create('Abstract virtual call.');
end;

constructor grep_setTok.Create;
  begin inherited Create; end;

function grep_setTok.match
  (c : char; caseIgnore : boolean) : boolean;
begin
  raise EgrepException.Create('Abstract virtual call.');
  result := false;
end;

procedure grep_setAlts.GRE_cat(GRE : cstr);
begin
  if (not Assigned(fseq)) or (not Assigned(GRE)) then exit;
  GRE.cat(fseq.str);
end;

constructor grep_setAlts.Create(const SEQ : PChar);
begin
  inherited Create;
  fseq := cstr.Create_PChar(SEQ);
end;

destructor grep_setAlts.Destroy;
begin
  fseq.Free;
  inherited Destroy;
end;

function grep_setAlts.match
    (c : char; caseIgnore : boolean) : boolean;
begin
    result := false;
    if Assigned(fseq) then
      if (strchr(fseq.str,c) <> nil) then
          result := true
      else if (caseIgnore) then
          if (isupper(c)) then begin
              if (strchr(fseq.str,tolower(c)) <> nil) then
                  result := true;
          end
          else if (strchr(fseq.str,toupper(c)) <> nil) then
              result := true;
end;

procedure grep_setRange.GRE_cat(GRE : cstr);
begin
  if not Assigned(GRE) then exit;
  GRE.ncat_char_fill(_left,1);
  GRE.cat('-').ncat_char_fill(_right,1);
end;

constructor grep_setRange.Create(left, right : char);
begin
  inherited Create;
  _left := left;
  _right := right;
  if Ord(_left) >= Ord(_right) then
     raise EgrepException.Create('Invalid grep set range.');
end;

function grep_setRange.match
    (c : char; caseIgnore : boolean) : boolean;
begin
    if (caseIgnore) then
        result := (Ord(toupper(c)) >= Ord(toupper(_left)))
            and (Ord(toupper(c)) <= Ord(toupper(_right)))
    else
        result := (Ord(c) >= Ord(_left)) and (Ord(c) <= Ord(_right));
end;

procedure grep_set.GRE_cpy;
var i : word;
begin
  if (not Assigned(fsetToks)) or (not Assigned(fGRE)) then exit;
  i := 0;
  pGRE.cpy('[');
  if negate then
     pGRE.cat('^');
  while i < fsetToks.nodes do begin
    grep_setTok(fsetToks[i]).GRE_cat(fGRE);
    inc(i);
    end;
  pGRE.cat(']');
  GRE_catCount;
end;

constructor grep_set.Create
    (const GRE : PChar; GRE_len : word; var tokLen : word);
var cspnLen, oldLen : word;
    alts : cstr;
begin
  inherited Create;
  negate := false;
  fsetToks := Container.Create_specs(grep_setTok,CL_ALL_DESCENDANTS,CL_FREE);
  if not Assigned(fsetToks) then exit;
  if (GRE = nil) or (GRE[0] <> '[') then
     raise EgrepException.Create('Invalid grep set.');
  tokLen := 1;
  if (GRE[1] = '^')  then begin
      negate := true;
      inc(tokLen);
  end;
  alts := cstr.Create;
  if not Assigned(alts) then exit;
try
  while (tokLen < GRE_len)  do begin
    cspnLen := strcspn(@(GRE[tokLen]),'-]\[()^*+?|.{}');
    oldLen := alts.len;
    alts.ncat_extra(@(GRE[tokLen]),cspnLen,1);
    if alts.len <> cspnLen + oldLen then begin
       tokLen := 0;
       exit;
       end;
    inc(tokLen,cspnLen);
    if (tokLen >= GRE_len) then begin
      error(GRE,tokLen,'] not found');
      tokLen := 0; exit;
      end;
    if (strspn(@(GRE[tokLen]),'[()^*+?|.{}') > 0) then begin
      error(GRE,tokLen,'illegal grep set char');
      tokLen := 0; exit;
      end;
    case (GRE[tokLen]) of
      '-': begin
        if (alts.len = 0) then begin
          error(GRE,tokLen,'left range missing');
          tokLen := 0;
          exit;
        end;
        inc(tokLen);
        if (tokLen >= GRE_len)  then begin
          error(GRE,tokLen,'right range missing');
          tokLen := 0;
          exit;
        end;
        if (GRE[tokLen] = '\')  then begin
          inc(tokLen);
          if (tokLen >= GRE_len)  then begin
            error(GRE,tokLen,'char following \ missing');
            tokLen := 0;
            exit;
          end;
        end;
        if (not fsetToks.insQFL(grep_setRange.Create
              (alts[alts.len-1],GRE[tokLen]))) then begin
           tokLen := 0;
           exit;
           end;
        inc(tokLen);
        alts.rmvTail(alts.len-1);
        if (alts.len = 0) then
            continue;
        { fall through for alts preceeding range }
        if (alts.len > 0) then
            if (not fsetToks.insQFL(grep_setAlts.Create(alts.str))) then begin
                tokLen := 0;
                exit;
                end;
        inc(tokLen);
        alts.clr;
        if (fsetToks.nodes = 0) then begin
          error(GRE,tokLen,'nothing in set');
          tokLen := 0;
          exit;
        end;
        exit;
        end;
      ']': begin
        if (alts.len > 0) then
            if (not fsetToks.insQFL(grep_setAlts.Create(alts.str))) then begin
                tokLen := 0;
                exit;
                end;
        inc(tokLen);
        alts.clr;
        if (fsetToks.nodes = 0) then begin
          error(GRE,tokLen,'nothing in set');
          tokLen := 0;
          exit;
        end;
        exit;
        end;
      '\': begin
        inc(tokLen);
        if (tokLen < GRE_len) then begin
          alts.ncat(@(GRE[tokLen]),1);
          inc(tokLen)
          end
        else begin
          error(GRE,tokLen,'nothing following \');
          tokLen := 0;
          exit;
        end;
        end;
      else raise EgrepException.Create('Invalid grep set.');
    end;
  end;
  tokLen := 0;
 finally
   alts.Free;
 end;
end;

destructor grep_set.Destroy;
begin
  fsetToks.Free;
  inherited Destroy;
end;

function grep_set.matchOne(const SS : PChar;
    var len : word; caseIgnore : boolean;
    groupToks : Container) : boolean;
var i : word;
begin
  { No group tokens will be matched within a set! }
    result := false;
    if not Assigned(fsetToks) then exit;
    if (negate)  then begin
      for i := 0 to fsetToks.nodes - 1 do
        if (not (fsetToks.atGet(i) as grep_setTok).match(SS[0],caseIgnore))
            then begin
            len := 1;
            result := true;
            exit;
            end;
    end
    else
      for i := 0 to fsetToks.nodes - 1 do
        if ((fsetToks.atGet(i) as grep_setTok).match(SS[0],caseIgnore))
            then begin
            len := 1;
            result := true;
            exit;
            end;
end;

procedure grep_alt.GRE_cpy;
begin
  if (not Assigned(_left)) or (not Assigned(_right)) or (not Assigned(fGRE))
     then exit;
  pGRE.cpy(_left.pGRE.str);
  pGRE.cat('|');
  pGRE.cat(_right.pGRE.str);
end;

constructor grep_alt.Create(const GRE : PChar;
    GRE_len : word; leftTok : grep_tok;
    var rightTokLen : word; groups : boolean);
begin
  if not Assigned(leftTok) then
     raise EgrepException.Create('Left token of alt missing.');
  inherited Create;
  _left := leftTok;
  _right := grep_tok.newTok(GRE,GRE_len,rightTokLen,groups);
end;

destructor grep_alt.Destroy;
begin
  _left.Free;
  _right.Free;
  inherited Destroy;
end;

function grep_alt.matchOne(const SS : PChar;
    var len : word; caseIgnore : boolean;
    groupToks : Container) : boolean;
begin
  { count is always One for grep_alt! }
  result := false;
  if not Assigned(_left) then
     raise EgrepException.Create('Left alt token missing.');
  if not Assigned(_right) then
     raise EgrepException.Create('Right alt token missing.');
  if (_left.matchOne(SS,len,caseIgnore,groupToks)) then
      result := true
  else if (_right.matchOne(SS,len,caseIgnore,groupToks)) then
      result := true;
  traceMatchOne(SS,result);
end;

procedure CLgrepExit; far;
begin
  ExitProc := ExitSave;
  MATCH_TRACE_OFF;
end;

begin
  ExitSave := ExitProc;
  ExitProc := @CLgrepExit;
  traceCstr := nil;
  NTRACE := true;
  grep_exp_sema4 := 0;
end.